#
#  PianoRay
#  Piano performance visualizer.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import ctypes
import json
import os
from typing import Mapping

import numpy as np
from tqdm import trange

from .. import logger
from ..cpp import build_lib, Types
from ..effects import render_blocks
from ..effects import parse_midi
from ..settings import Settings
from .video import Video


def preprocess(settings: Settings):
    """
    Does stuff to settings, e.g. change coords to pixels.
    """
    coord = settings.video.resolution[0] / 52
    settings.blocks.radius *= coord


def load_libs(cache: str) -> Mapping[str, ctypes.CDLL]:
    """
    Load libraries.
    """
    cache = os.path.join(cache, "c_libs")
    os.makedirs(cache, exist_ok=True)

    blocks = build_lib(
        ["effects/blocks/blocks.cpp"],
        cache,
        "blocks",
    )
    blocks.render_blocks.argtypes = [
        Types.img, Types.int, Types.int,
        Types.int,
        Types.int, Types.arr_int, Types.arr_double, Types.arr_double,
        Types.int, Types.double, Types.double, Types.double, Types.arr_uchar,
    ]

    return {
        "blocks": blocks,
    }


def render_video(settings: Settings, out: str, cache: str) -> None:
    preprocess(settings)
    libs = load_libs(cache)

    cache_settings = os.path.join(cache, "settings.json")
    cache_curr = os.path.join(cache, "currently_rendering.txt")

    # Check if continue previous render
    real_start = None
    if os.path.isfile(cache_settings) and os.path.isfile(cache_curr):
        with open(cache_settings, "r") as fp:
            prev_sets = Settings(json.load(fp))

        if prev_sets == settings:
            with open(cache_curr, "r") as fp:
                real_start = int(fp.read())
            print("Last render has same settings as this render, "
                  f"and stopped at frame {real_start}.")
            if input("Continue last render? [Y/n] ").lower().strip() == "n":
                real_start = None

    os.makedirs(cache, exist_ok=True)
    with open(cache_settings, "w") as fp:
        json.dump(settings._json(), fp)

    video = Video(os.path.join(cache, "output"), settings.audio.path,
        settings.audio.start)
    num_frames = render_frames(settings, libs, video, cache, real_start)
    video.compile(out, settings.video.fps, num_frames,
        settings.composition.margin_start, settings.video.vcodec)

    if os.path.isfile(cache_curr):
        os.remove(cache_curr)


def render_frames(settings, libs, video, cache, real_start=None) -> int:
    """
    Render frames.

    :param real_start: If applicable, currently_rendering
    :return: Number of frames rendered.
    """
    notes = parse_midi(settings)
    duration = int(max(x[3] for x in notes))

    fps = settings.video.fps
    m_start = settings.composition.margin_start
    m_end = settings.composition.margin_end
    frame_start = -fps * m_start
    frame_end = duration + fps*m_end

    num_frames = 0
    for frame in trange(frame_start, frame_end):
        num_frames += 1
        with open(os.path.join(cache, "currently_rendering.txt"), "w") as fp:
            fp.write(str(frame))

        if real_start is not None and frame < real_start:
            continue

        img = np.zeros((*settings.video.resolution[::-1], 3), dtype=np.uint8)
        render_blocks(libs["blocks"], settings, img, notes, frame)
        video.write(img)

    return num_frames



#
#  PianoRay
#  Video rendering pipeline with piano visualization.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import numpy as np

from ...cpp import Types


def render_blocks(lib, settings, img: np.ndarray, notes, frame: int):
    """
    Render the blocks.

    :param lib: C library for blocks.
    :param notes: MIDI notes from parse_midi.
    """
    keys = np.array([n[0] for n in notes], dtype=Types.int)
    starts = np.array([n[2] for n in notes], dtype=Types.double)
    ends = np.array([n[3] for n in notes], dtype=Types.double)

    settings_args = [settings.video.fps, settings.blocks.speed,
        settings.piano.black_width_fac, settings.blocks.radius,
        np.array(settings.blocks.color, dtype=np.uint8)]

    lib.render_blocks(
        img, img.shape[1], img.shape[0],
        frame,
        len(notes), keys, starts, ends,
        *settings_args,
    )



#
#  PianoRay
#  Video rendering pipeline with piano visualization.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from typing import Tuple

import numpy as np

from ..settings import Settings


def bounds(v, vmin, vmax):
    """
    Bound v between vmin and vmax.
    """
    return min(max(v, vmin), vmax)


def is_white_key(key: int) -> bool:
    """
    If the key is a white key on piano.
    """
    return (key % 12) not in (1, 4, 6, 9, 11)

def key_pos(key: int) -> float:
    """
    Position of the center of the key on the keyboard.
    Factor from 0 to 1 (start of first key to end of last).
    """
    white_width = 1 / 52

    last_white = False  # Last key was white
    pos = 0
    for k in range(key+1):
        white = is_white_key(k)
        if white:
            if last_white:
                pos += white_width
            else:
                pos += white_width / 2
        else:
            pos += white_width / 2
        last_white = white

    return pos

def key_coords(settings: Settings, key: int) -> Tuple[float, float]:
    """
    Horizontal (x) coordinates of key on the screen.

    :param key: Key.
    :return: ``(start_coord, end_coord)`` of key.
    """
    center = np.interp(key_pos(key), (0, 1), (0, settings.video.resolution[0]))
    white_width = settings.video.resolution[0] / 52
    black_width = white_width * settings.piano.black_width_fac
    width = white_width if is_white_key(key) else black_width
    half = width / 2
    return (center-half, center+half)


def note_coords(settings: Settings, event_frame: float,
        frame: float) -> float:
    """
    Vertical (y) coordinates of an event as it is dropping from the top
    to the keyboard.

    :param event_frame: The time of the event.
    :param frame: Current frame.
    :return: Y pixel position.
    """
    height = settings.video.resolution[1] / 2
    speed = (settings.blocks.speed * height / settings.video.fps)
    delta = speed * (frame-event_frame)
    return height + delta



#
#  PianoRay
#  Piano performance visualizer.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import os
import shutil

VERSION = "0.0.3"

PARENT = os.path.dirname(os.path.abspath(__file__))

FFMPEG = shutil.which("ffmpeg")
GCC = shutil.which("g++")
assert FFMPEG is not None
assert GCC is not None


SETTINGS_DEFAULT = {
    "video": {
        "fps": 30,
        "resolution": (1920, 1080),
        "vcodec": "libx265",
        "duration": None,
    },
    "audio": {
        "path": None,
        "start": 0,
    },
    "composition": {
        "margin_start": 2,
        "margin_end": 2,
    },
    "piano": {
        "black_width_fac": 0.6,
    },
    "blocks": {
        "speed": 0.5,
        "color": [185, 200, 255],
        "radius": 0.25,
    },
    "midi": {
        "file": None,
        "speed": 1,
    },
}


def forever():
    """
    Generator for indefinite tqdm progressbar.
    """
    while True:
        yield



#
#  PianoRay
#  Piano performance visualizer.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import os
import shutil
from subprocess import Popen, PIPE
from typing import Sequence

import cv2
import numpy as np

from .. import logger
from ..utils import FFMPEG


class Video:
    """
    Video.

    Saves frames to given cache directory and uses ffmpeg to make the
    final video.
    """

    def __init__(self, cache: str, audio: str = None,
            audio_offset: float = 0) -> None:
        """
        Initializes video.

        :param cache: Cache directory. Frames stored there.
        :param audio: Path to audio file.
        :param audio_offset: Seconds. Positive values play audio later.
        """
        self.cache = cache
        self.audio = audio
        self.audio_offset = audio_offset
        self.frame = 0

        os.makedirs(os.path.join(self.cache, "frames"), exist_ok=True)

    def write(self, img: np.ndarray) -> int:
        """
        Write a frame.
        Converts from RGB to BGR.

        :param img: Frame of shape ``(height, width, 3)``
        :return: This frame number.
        """
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        path = os.path.join(self.cache, "frames", str(self.frame)+".jpg")
        cv2.imwrite(path, img)

        self.frame += 1
        return self.frame - 1

    def compile(self, out: str, fps: int, num_frames: int,
            margin_start: float, vcodec="libx265", crf=24) -> None:
        """
        Use ffmpeg to compile frames and audio to video.

        :param out: Output video path.
        :param fps: Frames per second.
        :param num_frames: Total number of frames to compile.
        :param margin_start: settings.composition.margin_start
        :param vcodec: Video codec. Use libx264 if libx265 fails.
        :param crf: Constant rate factor. Higher values produce smaller
            file sizes but lower quality.
        """
        # Frames to video
        logger.info("Compiling frames to video.")
        args = [
            FFMPEG,
            "-y",
            "-i", os.path.join(self.cache, "frames", "%d.jpg"),
            "-vframes", num_frames,
            "-c:v", vcodec,
            "-crf", crf,
            "-r", fps,
            os.path.join(self.cache, "no_audio.mp4"),
        ]
        run_ffmpeg(args)

        if self.audio is not None:
            # Cut audio
            logger.info("Processing audio.")
            args = [
                FFMPEG,
                "-y",
                "-ss", self.audio_offset-margin_start, "-t", 100000,
                "-i", self.audio,
                os.path.join(self.cache, "offset.mp3"),
            ]
            run_ffmpeg(args)

            # Mix
            logger.info("Combining audio and video.")
            args = [
                FFMPEG,
                "-y",
                "-i", os.path.join(self.cache, "no_audio.mp4"),
                "-i", os.path.join(self.cache, "offset.mp3"),
                "-c", "copy",
                out,
            ]

            run_ffmpeg(args)

        else:
            # Copy to output.
            shutil.copy(os.path.join(self.cache, "no_audio.mp4"), out)

        logger.info(f"Video saved to {out}")


def run_ffmpeg(args: Sequence[str]):
    args = list(map(str, args))
    proc = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    proc.wait()

    if (code := proc.returncode) != 0:
        msg = f"FFmpeg exited with code {code}. " + \
            "Command: " + " ".join(args)
        raise ValueError(msg)



#
#  PianoRay
#  Piano performance visualizer.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from typing import Any, Mapping


class Settings:
    """
    Settings.
    """

    _data: Mapping[str, Any]

    def __init__(self, data: Mapping[str, Any]) -> None:
        object.__setattr__(self, "_data", {})
        for key in data:
            obj = data[key]
            if isinstance(obj, dict):
                obj = Settings(obj)
            self._data[key] = obj

    def __iter__(self):
        return iter(self._data)

    def __getattr__(self, name: str) -> Any:
        if name in self._data:
            return self._data[name]
        raise AttributeError(f"Settings object has no attribute {name}")

    def __setattr__(self, name: str, value: Any) -> None:
        self._data[name] = value

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Settings):
            return False
        return self._data == other._data

    def _merge(self, other: "Settings") -> None:
        """
        Merge other settings with self, keeping self if conflict.
        """
        for key in other:
            obj = getattr(other, key)
            if not hasattr(self, key):
                self._data[key] = obj
            else:
                if isinstance(obj, Settings):
                    getattr(self, key)._merge(obj)

    def _json(self) -> Mapping[str, Any]:
        ret = {}
        for key in self._data:
            obj = self._data[key]
            if isinstance(obj, Settings):
                obj = obj._json()
            ret[key] = obj

        return ret



#
#  PianoRay
#  Piano performance visualizer.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import time

import pygame

from .. import logger
from .video import Video

pygame.init()


def view_video(path: str) -> None:
    video = Video(path)
    logger.info(f"Extracting frames to {video.tmpdir}")
    logger.warn("This feature is not complete yet.")

    resized = False  # Redraw if resized
    display = pygame.display.set_mode((1280, 720), pygame.RESIZABLE)
    pygame.display.set_caption("PianoRay Viewer")

    run = True
    while run:
        time.sleep(1/30)
        pygame.display.update()
        events = pygame.event.get()

        for event in events:
            if event.type == pygame.QUIT:
                run = False

            elif event.type == pygame.VIDEORESIZE:
                resized = True

        display.fill((0, 0, 0))

    pygame.quit()
    video.run = False



#
#  PianoRay
#  Piano performance visualizer.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import argparse
import json
import os
import sys

from . import logger
from .settings import Settings
from .utils import SETTINGS_DEFAULT, VERSION

from .render import render_video
from .view import view_video


def render(args):
    if args.settings is None:
        logger.error("Please provide an argument for settings.")
        return 1
    if args.output is None:
        logger.error("Please provide an argument for output.")
        return 1

    if os.path.isfile(args.output) and not args.yes:
        print(f"Output file {args.output} already exists.")
        if input("Overwrite file? [y/N] ").lower().strip() != "y":
            return 3

    with open(args.settings, "r") as fp:
        settings = Settings(json.load(fp))
    default = Settings(SETTINGS_DEFAULT)
    settings._merge(default)

    render_video(settings, args.output, args.cache)

    if args.preview:
        os.system(f"xdg-open {args.output}")


def view(args):
    if len(args.options) == 0:
        logger.error("Please provide the video file to view.")
        return 1

    view_video(args.options[0])


def main():
    parser = argparse.ArgumentParser(
        description="Piano performance visualizer.")
    parser.add_argument("-V", "--version", action="version",
        help="Show version info.", version=f"Pianoray v{VERSION}")
    parser.add_argument("-s", "--settings",
        help="Path to settings JSON file.")
    parser.add_argument("-o", "--output",
        help="Output video file.")
    parser.add_argument("-c", "--cache",
        help="Cache path (default .prcache)", default=".prcache")
    parser.add_argument("-p", "--preview", action="store_true",
        help="Open output file after rendering")
    parser.add_argument("-y", "--yes",
        help="Don't prompt overwrite.", action="store_true")
    parser.add_argument("mode", choices={"render", "view"}, nargs="?",
        help="Mode to run.")
    parser.add_argument("options", nargs="*",
        help="Any other positional arguments.")
    args = parser.parse_args()

    if args.mode is None or args.mode == "render":
        return render(args)
    elif args.mode == "view":
        return view(args)


if __name__ == "__main__":
    sys.exit(main())



#
#  PianoRay
#  Piano performance visualizer.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import os
import random
from threading import Thread

import cv2
import numpy as np

from .utils import TMP


class Video:
    """
    Handles extracting and managing frames.
    Saves to tmp directory and extracts in the background.
    """

    run: bool
    tmpdir: str
    num_frames: int
    extracted: int

    def __init__(self, path: str):
        """
        :param path: Path to video.
        """
        self.run = True

        rand = str(random.randint(0, 10000000))
        self.tmpdir = os.path.join(TMP, f"pianoray_viewer_{rand}")
        os.makedirs(self.tmpdir, exist_ok=True)

        self._video = cv2.VideoCapture(path)
        self.num_frames = int(self._video.get(cv2.CAP_PROP_FRAME_COUNT))

        Thread(target=self.extract).start()

    def get(self, frame: int) -> np.ndarray:
        """
        Get the frame image.

        :return: np array if successful, None else.
        """
        if frame > self.extracted:
            return None

        path = os.path.join(self.tmpdir, f"{frame}.jpg")
        return cv2.imread(path)

    def extract(self):
        """
        Extract all the frames.
        Run this in a different thread.
        Set self.run to False to stop.
        """
        for frame in range(self.num_frames):
            if not self.run:
                break
            ret, img = self._video.read()
            if not ret:
                break

            path = os.path.join(self.tmpdir, f"{frame}.jpg")
            cv2.imwrite(path, img)

            self.extracted = frame



#
#  PianoRay
#  Piano performance visualizer.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import os
from typing import Sequence, Tuple

import mido

from ..settings import Settings


def parse_midi(settings: Settings) -> Sequence[Tuple[float, float]]:
    """
    Parse midi file.

    :param settings: Settings.
    :return: List of ``(note, velocity, start_frame, end_frame)``.
    """
    path = settings.midi.file
    if path is None or not os.path.isfile(path):
        raise ValueError("Setting midi.file invalid.")

    midi = mido.MidiFile(path)
    notes = []

    t = 0
    started = False
    starts = [0] * 255
    for msg in midi:
        if started:  # Only increment after first note
            t += msg.time * settings.video.fps / settings.midi.speed

        if msg.type.startswith("note_"):
            started = True

            vel = msg.velocity
            note = msg.note - 21
            if msg.type == "note_on" and vel > 0:
                starts[note] = t
            else:
                notes.append((note, vel, starts[note], t))

    return notes



#
#  PianoRay
#  Video rendering pipeline with piano visualization.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import sys
import termcolor
from datetime import datetime


def time():
    now = datetime.now()
    return now.strftime("%H:%M:%S")


def log(type: str, msg: str, color: str):
    s = f"[{time()}] {type}:"
    s += " " * (6-len(type))
    s += msg
    print(termcolor.colored(s, color), file=sys.stderr)

def info(msg: str) -> None:
    """
    Info log to stderr.
    Color: blue
    """
    log("INFO", msg, "blue")

def warn(msg: str) -> None:
    """
    Warning log to stderr.
    Color: yellow
    """
    log("WARN", msg, "yellow")

def error(msg: str) -> None:
    """
    Error log to stderr.
    Color: red
    """
    log("ERROR", msg, "red")



#
#  PianoRay
#  Piano performance visualizer.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import os
import setuptools

VERSION = os.environ.get("PYPI_VERSION", "0.0.3")
VERSION = VERSION.split("/")[-1]  # GitHub input


with open("README.md", "r") as fp:
    long_description = fp.read()

with open("requirements.txt", "r") as fp:
    requirements = fp.read().strip().split("\n")

setuptools.setup(
    name="pianoray",
    version=VERSION,
    author="PianoRay Authors",
    author_email="phuang1024@gmail.com",
    description="Piano performance visualizer.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/phuang1024/pianoray",
    py_modules=["pianoray"],
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "pianoray = pianoray.main:main",
        ]
    },
)



#
#  PianoRay
#  Video rendering pipeline with piano visualization.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import ctypes
import os
from subprocess import Popen
from typing import Sequence

import numpy as np
from numpy.ctypeslib import ndpointer

from . import logger
from .utils import GCC

PARENT = os.path.dirname(os.path.abspath(__file__))

CPP_UTILS = os.path.join(PARENT, "cutils")
CPP_UTILS_FILES = [os.path.join(CPP_UTILS, f)
    for f in os.listdir(CPP_UTILS) if f.endswith(".cpp")]


class Types:
    _arr_flags = "aligned, c_contiguous"

    char = ctypes.c_int8
    uchar = ctypes.c_uint8
    int = ctypes.c_int32
    uint = ctypes.c_uint32
    float = ctypes.c_float
    double = ctypes.c_double

    arr_uchar = ndpointer(dtype=uchar, ndim=1, flags=_arr_flags)
    arr_int = ndpointer(dtype=int, ndim=1, flags=_arr_flags)
    arr_double = ndpointer(dtype=double, ndim=1, flags=_arr_flags)

    img = ndpointer(dtype=uchar, ndim=3, flags=_arr_flags)


def build_lib(files: Sequence[str], cache: str, name: str) -> ctypes.CDLL:
    """
    Initialize the lib.

    :param files: C files relative to THIS file.
    :param cache: Cache directory.
    :param name: Name of the library.
    :return: C library.
    """
    logger.info(f"Building C library {name}")

    cache = os.path.join(cache, name)
    os.makedirs(cache, exist_ok=True)

    files = [os.path.join(PARENT, f) for f in files]
    files.extend(CPP_UTILS_FILES)

    obj_files = []
    for f in files:
        fname = os.path.basename(f)
        obj_name = os.path.splitext(fname)[0] + ".o"
        obj_path = os.path.join(cache, obj_name)
        obj_files.append(obj_path)
        compile(f, obj_path)

    lib_path = os.path.join(cache, f"lib{name}.so")
    link(obj_files, lib_path)

    return ctypes.CDLL(lib_path)

def compile(cpp, obj):
    args = [GCC, "-Wall", "-O3", "-c", "-fPIC", cpp, "-o", obj, "-I", CPP_UTILS]
    p = Popen(args)
    p.wait()
    assert p.returncode == 0

def link(obj_files, lib_path):
    args = [GCC, "-shared", "-o", lib_path, *obj_files]
    p = Popen(args)
    p.wait()
    assert p.returncode == 0
